import os
import unittest
import json
import logging

# external imports
from chainlib.eth.constant import ZERO_ADDRESS
from chainlib.eth.nonce import RPCNonceOracle
from chainlib.eth.tx import receipt
from chainlib.eth.block import (
        block_latest,
        block_by_number,
        )
from chainlib.eth.address import to_checksum_address
from hexathon import (
        strip_0x,
        add_0x,
        )

# local imports
from erc20_demurrage_token import DemurrageToken

# test imports
from tests.base import TestDemurrageUnit

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()

testdir = os.path.dirname(__file__)


class TestRedistribution(TestDemurrageUnit):

    def test_single_step(self):
        nonce_oracle = RPCNonceOracle(self.accounts[0], self.rpc)
        c = DemurrageToken(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)

        mint_amount = 100000000

        (tx_hash, o) = c.mint_to(self.address, self.accounts[0], self.accounts[1], mint_amount)
        self.rpc.do(o)

        self.backend.time_travel(self.start_time + self.period_seconds)

        (tx_hash, o) = c.change_period(self.address, self.accounts[0])
        self.rpc.do(o)

        expected_balance = int(mint_amount - ((self.tax_level / 1000000) * mint_amount))

        o = c.balance_of(self.address, self.accounts[1], sender_address=self.accounts[0])
        r = self.rpc.do(o)
        balance = c.parse_balance_of(r)

        self.assertEqual(balance, expected_balance)


    def test_single_step_multi(self):
        nonce_oracle = RPCNonceOracle(self.accounts[0], self.rpc)
        c = DemurrageToken(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)

        mint_amount = 100000000

        for i in range(3):
            (tx_hash, o) = c.mint_to(self.address, self.accounts[0], self.accounts[i+1], mint_amount)
            self.rpc.do(o)
    
        self.backend.time_travel(self.start_time + self.period_seconds)

        (tx_hash, o) = c.change_period(self.address, self.accounts[0])
        self.rpc.do(o)

        expected_balance = int(mint_amount - ((self.tax_level / 1000000) * mint_amount))

        for i in range(3):
            o = c.balance_of(self.address, self.accounts[i+1], sender_address=self.accounts[0])
            r = self.rpc.do(o)
            balance = c.parse_balance_of(r)
            self.assertEqual(balance, expected_balance)


    def test_single_step_transfer(self):
        nonce_oracle = RPCNonceOracle(self.accounts[0], self.rpc)
        c = DemurrageToken(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)

        mint_amount = 100000000
        half_mint_amount = int(mint_amount / 2)

        (tx_hash, o) = c.mint_to(self.address, self.accounts[0], self.accounts[1], mint_amount)
        self.rpc.do(o)

        (tx_hash, o) = c.mint_to(self.address, self.accounts[0], self.accounts[2], mint_amount)
        self.rpc.do(o)
    
        nonce_oracle = RPCNonceOracle(self.accounts[1], self.rpc)
        c = DemurrageToken(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash, o) = c.transfer(self.address, self.accounts[1], self.accounts[3], half_mint_amount)
        self.rpc.do(o)

        self.backend.time_travel(self.start_time + self.period_seconds)

        (tx_hash, o) = c.change_period(self.address, self.accounts[1])
        self.rpc.do(o)

        demurrage_amount = int((self.tax_level / 1000000) * mint_amount)

        expected_balance = mint_amount - demurrage_amount
        o = c.balance_of(self.address, self.accounts[2], sender_address=self.accounts[0])
        r = self.rpc.do(o)
        balance = c.parse_balance_of(r)
        self.assertEqual(balance, expected_balance)

        half_demurrage_amount = int((self.tax_level / 1000000) * half_mint_amount)

        expected_balance = half_mint_amount - half_demurrage_amount
        o = c.balance_of(self.address, self.accounts[1], sender_address=self.accounts[0])
        r = self.rpc.do(o)
        balance = c.parse_balance_of(r)
        self.assertEqual(balance, expected_balance)

        o = c.balance_of(self.address, self.accounts[3], sender_address=self.accounts[0])
        r = self.rpc.do(o)
        balance = c.parse_balance_of(r)
        self.assertEqual(balance, expected_balance)

        o = c.total_supply(self.address, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        supply = c.parse_total_supply(r)

        expected_balance = int(supply * (self.tax_level / 1000000))

        o = c.balance_of(self.address, self.sink_address, sender_address=self.accounts[0])
        r = self.rpc.do(o)
        balance = c.parse_balance_of(r)
        self.assertEqual(balance, expected_balance)


if __name__ == '__main__':
    unittest.main()
